# SPDX-License-Identifier: MPL-2.0
"""
# baseAnything: Encode/decode any integer into/from a series of symbols of a custom palette.

#### Basic use
```
>>> bxbinary = baseX('01')  # Instantiates a baseX object with a Base-2 aka binary string-symbol alphabet
>>> bxbinary % 7
'111'

```

#### Proof: A roundtrip
```
>>> bxhexadecimal = baseX('0123456789ABCDEF')  # Base-16 aka hexadecimal
>>> 9000 == int(bxhexadecimal % 9000, base=16)
True

```

#### Arbitrary symbols
One can use arbitrary symbols. Anything. `['bla', 2, object(), object()]` is a valid symbol alphabet.

It may be desirable to get a symbol list back rather than a string, so that the symbol framing is kept
intact — because, for instance, a symbol alphabet of `['a', 'aa']` yields ambiguity in its output when it's
mashed into a string; `'aaa'` has 3 possible interpretations.

The `%` operator yields strings when encoding an integer.
The `/` operator makes a baseX object return the encoding as a list of symbols instead.

```
>>> bxfunky = baseX(['Do', 'Re', 'Mi', 'Fa', 'Sol', 'La', 'Ti'])
>>> bxfunky % 10  # Emits a warning, because the result may not be decodable!
'ReFa'
>>> bxfunky / 10  # Symbols returned as a list rather than concatenated into a string.
['Re', 'Fa']
>>> 10 == bxfunky / (bxfunky / 10)
True

```

#### Bizarre encodings
Speaking of arbitrary symbols... how about this visually compact encoding:

```
>>> bxbizarre = baseX(list(map(chr, range(0x0300, 0x036f))))  # unicode diacritical marks
>>> wut = 'O' + bxbizarre % 1234567890
>>> wut
'Ơ̡͎̈̎'
>>> bxbizarre / wut[1:]
1234567890

```
"""


from typing import Sequence, Mapping
from collections import Counter
from warnings import warn


def enbaseX(num: int, alphabet: Sequence, exponent: int = 1):
    if num == 0 and exponent > 1: return []
    base = len(alphabet)
    magnitude = base ** exponent
    chomp = num % magnitude
    sym = alphabet[int(chomp / (magnitude / base))]
    remaining = num - chomp
    if num < base: return [sym]
    return enbaseX(remaining, alphabet, exponent + 1) + [sym]


def debaseX(input: Sequence, alphabet: Mapping, base: int = 1):
    if base == 1 and len(input) == 0: raise ValueError("Can't interpret a zero-length input")
    if not len(input): return 0
    terminus = input[-1]
    return debaseX(input[:-1], alphabet, base * len(alphabet)) + (alphabet[terminus] * base)


class baseX:

    def __init__(self, alphabet: Sequence):
        if len(alphabet) < 2:
            raise ValueError("Not enough symbols in alphabet. You need at least 2 symbols.")
        sym, cnt = Counter(alphabet).most_common(1)[0]
        if cnt > 1:
            raise ValueError(f'Non-unique symbol in alphabet: "{sym}"')
        self.alphabet = alphabet
        self.mapping = dict((sym, i) for (i, sym) in enumerate(alphabet))
        self.safetostringjoin = all(map(lambda thing: isinstance(thing, str) and len(thing) == 1, alphabet))

    def encode(self, num: int):
        return enbaseX(num, self.alphabet)

    def decode(self, input: Sequence):
        return debaseX(input, self.mapping)

    def __mod__(self, other):
        if isinstance(other, int):
            if not self.safetostringjoin:
                warn(
                    "Your alphabet doesn't look safe to use the string-joining shortcut with; output may not be decodable. Use the '/' operator instead of the '%' operator.",
                    stacklevel=2,
                )
            return ''.join(map(str, self.encode(other)))
        return self.__truediv__(other)


    def __truediv__(self, other):
        if isinstance(other, int):
            return self.encode(other)
        if isinstance(other, Sequence):
            return self.decode(other)
        raise ValueError(f'Not an integer (to encode) or sequence (to decode): "{repr(other)}"')

    def __repr__(self):
        return f'baseX({repr(self.alphabet)})'

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.alphabet == other.alphabet

    def __len__(self):
        return len(self.alphabet)